local addon, shared = ...

local NODEBUG=		99
local WDOGYIELD=	15
local FUNCTIONRUN=	10
local STATUSUPDT=	8
local INFORMATION=	5
local DETAIL=		1

local TRK_INACTIVE=	0
local TRK_START=	1
local TRK_ACTIVE=	2
local TRK_FINISHED=	3

local trk_status = TRK_INACTIVE

local sqrt = math.sqrt
local atan = math.atan2

local last_dir = "N"
local last_dist = 0

---------------------
-- Variables
---------------------

_WayPoint = {}
local WP = _WayPoint

Waypoint_Settings = {}

WP.UIElements = {
	}
	
WP.UIElements.context = UI.CreateContext("Waypoint")

local default_settings = {
	textupdate = true, updatetime = 2,
	x = 500, y = 500,
	scale = 1, distance = 5, show = true, absolute = true
}

local wp_coords = {}
local prev_detail = nil
local tx_updatetext = default_settings.updatetime
local dest_distance = default_settings.distance

---------------------
-- Functions
---------------------

local debuglevel = NODEBUG

local function Debug(s,l)
	if (l or INFORMATION) >= debuglevel then
		print(s)
	end
end

local function MergeTable(o,n)
	for k,v in pairs(n) do
		if type(v) == "table" then
			if o[k] == nil then
				o[k] = {}				
			end
	 	 	if type(o[k]) == 'table' then
	 			MergeTable(o[k], n[k])
	 	 	end
		else
			if o[k] == nil then
				o[k] = v
			end
		end
	end
end

function WP.CreateUI()
	WP.UIElements.icon = UI.CreateFrame("Texture", "icon", WP.UIElements.context)
	WP.UIElements.icon:SetTexture(addon.identifier, "compass.png")
	WP.UIElements.icon:SetVisible(false)

	function WP.UIElements.icon.Event:LeftDown()
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - WP.UIElements.icon:GetLeft()
		self.sy = mouseData.y - WP.UIElements.icon:GetTop()
	end
	
	function WP.UIElements.icon.Event:MouseMove()
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			WP.UIElements.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end
	
	function WP.UIElements.icon.Event:LeftUp()
		if self.MouseDown then
			self.MouseDown = false
		end
		Waypoint_Settings.x = WP.UIElements.icon:GetLeft()
		Waypoint_Settings.y = WP.UIElements.icon:GetTop()
	end
	
	function WP.UIElements.icon.Event:RightClick()
		trk_status = TRK_FINISHED
	end

	WP.UIElements.header = UI.CreateFrame("Text", "header", WP.UIElements.icon)
	WP.UIElements.header:SetText(shared.translation["Waypoint"])
	WP.UIElements.header:SetPoint("TOPLEFT", WP.UIElements.icon, "TOPRIGHT", 0, 0)
	WP.UIElements.header:SetBackgroundColor(0,0,0,0.5)
	
	WP.UIElements.compass = UI.CreateFrame("Texture", "compass", WP.UIElements.icon)
	WP.UIElements.compass:SetTexture(addon.identifier, "compass-N.png")
	WP.UIElements.compass:SetPoint("TOPLEFT", WP.UIElements.icon, "BOTTOMLEFT")
	WP.UIElements.compass:SetVisible(false)
end

function WP.ScaleUI()
	local h32 = 32 * Waypoint_Settings.scale
	local h14 = 14 * Waypoint_Settings.scale
	local h96 = 96 * Waypoint_Settings.scale
	
	WP.UIElements.icon:SetWidth(h32)
	WP.UIElements.icon:SetHeight(h32)
	WP.UIElements.header:SetFontSize(h14)
	WP.UIElements.compass:SetWidth(h96)
	WP.UIElements.compass:SetHeight(h96)
end

local __cpxdirs = {
	"S", "SE", "E", "NE", "N", "NW", "W", "SW", "S",
	"South", "South-East", "East", "North-East", "North", "North-West", "West", "South-West", "South"
	}

function WP.direction(dt)

	if dt > 180 then dt = dt-360 end
	if dt < -180 then dt = dt+360 end

	local d_file = tonumber(string.format("%d", dt/15))
	if d_file == -12 then d_file = 12 end
	
	local d_comp = tonumber(string.format("%d", dt/45)) + 5
	
	local dr
	if d_file < 0 then
		dr=string.format("n%d", d_file*-15)
	elseif d_file > 0 then
		dr=string.format("p%d", d_file*15)
	else
		dr="0"
	end

	return {s=dr, l = shared.translation[__cpxdirs[d_comp]]}
--[[	
	if dt >= 157.5		then	d = "S"
	elseif dt >= 112.5	then	d = "SW"
	elseif dt >= 67.5	then	d = "W"
	elseif dt >= 22.5	then	d = "NW"
	elseif dt >= -22.5	then	d = "N"
	elseif dt >= -67.5	then	d = "NE"
	elseif dt >= -112.5	then	d = "E"
	elseif dt >= -157.5	then	d = "SE"
	else						d = "S"
	end
	
	return {s = d, l = shared.translation[d]}
	]]
end

function WP.UIUpdateCompass()
	local dtl = Inspect.Unit.Detail("player")
	if dtl.coordX ~= nil then
		local dx = dtl.coordX - wp_coords.x
		local dy = dtl.coordZ - wp_coords.y
		
		local dt = atan(dx,dy)*180/math.pi
		
		if Waypoint_Settings.absolute then
			local d = WP.direction(dt).s
			if d ~= last_dir then
				last_dir = d
				Debug("Loading texture: compass-"..d..".png", DETAIL)
				WP.UIElements.compass:SetTexture(addon.identifier, "compass-"..d..".png")
				WP.UIElements.compass:SetAlpha(1)
			end
		else
			if prev_detail then
				local p_dx = prev_detail.coordX - dtl.coordX
				local p_dy = prev_detail.coordZ - dtl.coordZ
				
				local dp = atan(p_dx,p_dy)*180/math.pi
				
				if dp ~= 0 then
					local dx = dt-dp
					if dx < -180 then dx = dx + 360
					elseif dx > 360 then dx = dx - 360
					end
					Debug(string.format("P: %d, T: %d, diff: %d", dp, dt, dx), DETAIL)
					local d = WP.direction(dx).s
					if d ~= last_dir then
						last_dir = d
						Debug("Loading texture: compass-"..d..".png", DETAIL)
						WP.UIElements.compass:SetTexture(addon.identifier, "compass-"..d..".png")
					end
					WP.UIElements.compass:SetAlpha(1)
				else
					WP.UIElements.compass:SetAlpha(0.25)
				end
				
			end
		end
		
		local d = sqrt((dx^2)+(dy^2))
		if d <= dest_distance then
			print(shared.translation["You have reached your destination!"])
			trk_status = TRK_FINISHED
		end
	end
	prev_detail = dtl
end

local skip_cnt = 0

function WP.UIPrintUpdate()
	if Waypoint_Settings.textupdate then
		local dtl = Inspect.Unit.Detail("player")
		if dtl.coordX ~= nil then
		
			local dx = dtl.coordX - wp_coords.x
			local dy = dtl.coordZ - wp_coords.y
		
			local dt = atan(dx,dy)*180/math.pi
			
			local d = sqrt((dx^2)+(dy^2))

			Debug(string.format("pX = %d, pY = %d, tX = %d, tY = %d", dtl.coordX, dtl.coordZ, wp_coords.x, wp_coords.y), DETAIL)
			
			if (d ~= last_dist) or (skip_cnt == 10) then
				print(string.format(shared.translation["Your target is %.2fm away, heading %s"], d, WP.direction(dt).l))
				last_dist = d
				skip_cnt = 0
			else
				skip_cnt = skip_cnt+1
			end
		end
	end
end

local last_update1s = 0
local last_update2s = 0

function WP.OnUpdate()
	if trk_status == TRK_START then
		last_update1s = 0
		last_update2s = 0
		WP.UIElements.icon:SetVisible(true)
		WP.UIElements.compass:SetVisible(true)
		trk_status = TRK_ACTIVE
	elseif trk_status == TRK_ACTIVE then
		local tx = Inspect.Time.Real()
		if tx - last_update1s > 0.5 then
			WP.UIUpdateCompass()
			--print("1s update")
			last_update1s = tx
		end
		if tx - last_update2s > tx_updatetxt then
			--print("2s update")
			WP.UIPrintUpdate()
			last_update2s = tx
		end
	elseif trk_status == TRK_FINISHED then
		WP.UIElements.compass:SetVisible(false)
		WP.UIElements.icon:SetVisible(Waypoint_Settings.show)
		WP.UIElements.header:SetText("Waypoint")
		trk_status = TRK_INACTIVE
	end
end

function WP.SVLoaded()
	Waypoint_Settings = Waypoint_Settings or {}
	MergeTable(Waypoint_Settings, default_settings)
	WP.ScaleUI()
	WP.UIElements.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", Waypoint_Settings.x, Waypoint_Settings.y)
	WP.UIElements.icon:SetVisible(Waypoint_Settings.show)
	tx_updatetxt = Waypoint_Settings.updatetime
	dest_distance = Waypoint_Settings.distance
end

function WP.ShowUsage()
	print(string.format("/wp X Y -- %s", shared.translation["sets your target waypoint to world coords X,Y"]))
	print(string.format("/wp %s -- %s", shared.translation["clear"], shared.translation["clears the current waypoint (or right-click the compass icon)"]))
	print(string.format("/wp %s -- %s", shared.translation["mode"], shared.translation["toggles between absolute and relative heading modes [default = Absolute]"]))
	print(string.format("/wp %s -- %s", shared.translation["hide"], shared.translation["toggles display of UI when not actively tracking [default = SHOWN]"]))
	print(string.format("/wp %s N -- %s", shared.translation["distance"], shared.translation["sets how close you need to get to target [default = 5]"]))
	print(string.format("/wp %s N -- %s", shared.translation["update"], shared.translation["sets text update time to N seconds [default = 2]"]))
	print(string.format("/wp %s -- %s", shared.translation["text"], shared.translation["toggles display of chat frame updates [default = ON]"]))
	print(string.format("/wp %s N.NN -- %s", shared.translation["scale"], shared.translation["resizes UI elements to N.NN times default"]))
end

function WP.SlashHandler(args)
	Debug("ARGS: "..args, DETAIL)
	local r = {}
	local numargs = 0
	for token in string.gmatch(args, "[^%s]+") do
		r[numargs] = token
		numargs=numargs+1
	end
	if numargs > 0 then
		if r[0] == shared.translation["scale"] then
			local sc = tonumber(r[1])
			if sc ~= nil then
				Waypoint_Settings.scale = sc
				print(string.format(shared.translation["Scaling UI to %s"], sc))
				WP.ScaleUI()
			else
				print(shared.translation["Invalid arguments"])
				WP.ShowUsage()
			end
		elseif r[0] == shared.translation["distance"] then
			local dx = tonumber(r[1])
			if dx ~= nil then
				dest_distance = dx
				Waypoint_Settings.distance = dx
				print(string.format(shared.translation["Setting destination distance to %dm"],dx))
			else
				print(shared.translation["Invalid arguments"])
				WP.ShowUsage()
			end
		elseif r[0] == shared.translation["update"] then
			local tx = tonumber(r[1])
			if tx ~= nil then
				tx_updatetxt = tx
				Waypoint_Settings.updatetime = tx
				print(string.format(shared.translation["Setting update text delay to %ds"],tx))
			else
				print(shared.translation["Invalid arguments"])
				WP.ShowUsage()
			end
		elseif r[0] == shared.translation["text"] then
			Waypoint_Settings.textupdate = not Waypoint_Settings.textupdate
			print(shared.translation["Text updates for direction: "]..tostring(Waypoint_Settings.textupdate))
		elseif r[0] == shared.translation["hide"] then
			print(string.format(shared.translation["Hiding UI when not tracking: %s"], tostring(Waypoint_Settings.show)))
			Waypoint_Settings.show = not Waypoint_Settings.show
			WP.UIElements.icon:SetVisible(Waypoint_Settings.show)
		elseif r[0] == shared.translation["mode"] then
			Waypoint_Settings.absolute = not Waypoint_Settings.absolute
			if Waypoint_Settings.absolute then
				print(shared.translation["Mode set to absolute. Arrows show compass direction of waypoint"])
			else
				print(shared.translation["Mode set to relative. Arrows show direction of waypoint relative to the heading of your character"])
			end
		elseif r[0] == shared.translation["clear"] then
			trk_status = TRK_FINISHED
		else
			local _coords = {}
			local cnt=1
			local d_args = string.gsub(args, "[^%d]", " ")
			for token in string.gmatch(d_args, "[%d]+") do
				_coords[cnt] = token
				cnt=cnt+1
			end
			
			local c1 = tonumber(_coords[1])
			local c2 = tonumber(_coords[2])
			
			Debug(string.format("a1: %s, a2: %s, c1: %s, c2: %s", tostring(_coords[1]), tostring(_coords[2]), tostring(c1), tostring(c2)), DETAIL)

			if (c1~=nil) and (c2~=nil) then
				WP.UIElements.header:SetText(string.format("%s: %d, %d", shared.translation["Waypoint"], c1,c2))
				wp_coords.x = c1
				wp_coords.y = c2
				print(string.format("%s %d, %d", shared.translation["Tracking to"], c1,c2))
				trk_status = TRK_START
			else
				print(string.format("%s: %s", shared.translation["Can make no sense of"], args))
				WP.ShowUsage()
			end
		end
	else
		WP.ShowUsage()
	end
end

---------------------
-- Events
---------------------

table.insert(Event.System.Update.Begin, { WP.OnUpdate, addon.identifier, "Event.System.Update.Begin" })
table.insert(Event.Addon.SavedVariables.Load.End, { WP.SVLoaded, addon.identifier, "Event.Addon.SavedVariables.Load.End" })
table.insert(Command.Slash.Register("wp"), { WP.SlashHandler, addon.identifier, "Command.Slash.Register" })
table.insert(Command.Slash.Register("waypoint"), { WP.SlashHandler, addon.identifier, "Command.Slash.Register" })

print(string.format("v%s loaded.", addon.toc.Version))

WP.CreateUI()